@@ -27,13 +27,14 @@ import math
 import getopt
 import sys
 
-version = "2.0"
+version = "2.1"
 verbose = 0 # 0 means not verbose, higher numbers show more information, see maybe_print for more info
 sec_loop_timeout = 200 # in ms, clocks are incremented every time this timeout ends
 ms_loop_timeout = 50 # in ms, the same as above, in milisecond mode
 default_time = 2 * 60 * 1000 # in ms
 default_fisher = 10 # in seconds, default is used only by the UI
 default_players = 2 # default number of players
+default_game = "BlitzGame" # the default game type. note that this cannot be FisherGame because of the way the UI is setup.
 
 class Clock:
     """The main clock engine
@@ -54,6 +55,8 @@ class Clock:
     game = None
     # a clock is part of a chained list
     next = None
+    # usually, clocks go backwards, but some games might require it to go other ways
+    factor = -1
 
     def __init__(self, game):
         """Setup the clock engine
@@ -92,8 +95,6 @@ class Clock:
         if self.last_start:
             self.time = self.get_time()
             self.last_start = 0
-            if self.game.fisher:
-                self.time += self.game.fisher
         self.update()
 
     def pause(self):
@@ -114,7 +115,7 @@ class Clock:
             diff = time.time() - self.last_start
         else:
             diff = 0
-        return self.time - diff*1000
+        return self.time + (self.factor * diff*1000)
 
     def update(self):
         """Refresh the display of the clock's widget"""
@@ -144,11 +145,12 @@ class Game:
     """the game engine
     
     this regroups two clocks and handles turn switches
+
+    to simplify the code logic, this is basically the "Blitz" mode
     """
 
     # the number of turns played
     turns = 0
-    fisher = 0 # 0 if fisher disabled, otherwise in ms
     miliseconds = False
     mode = 'blitz'
     
@@ -158,9 +160,8 @@ class Game:
     cur_clock = None
     first_clock = None
 
-    def __init__(self, clocks, miliseconds, fisher):
+    def __init__(self, clocks, miliseconds):
         self.miliseconds = miliseconds
-        self.fisher = fisher
 
         # the clock engines
         self.cur_clock = self.first_clock = p = Clock(self)
@@ -225,7 +226,94 @@ class Game:
         
         we basically dump all variables and some functions
         """
-        return "  game engine %s\n  turns: %f / %d fisher: %d miliseconds: %d\n  first %s\n  current %s" % ( object.__str__(self), self.turns, self.get_turns(), self.fisher, self.miliseconds, self.first_clock, self.cur_clock)
+        return "  game engine %s\n  turns: %f / %d miliseconds: %d\n  first %s\n  current %s" % ( object.__str__(self), self.turns, self.get_turns(), self.miliseconds, self.first_clock, self.cur_clock)
+
+class BlitzGame(Game):
+    """a regular, blitz-style chess game. note that this is just an alias to the base Game class which implements everything"""
+    pass
+
+class FisherGame(Game):
+    """a regular, blitz-style chess game"""
+    fisher = 0
+
+    def __init__(self, clocks, miliseconds, fisher):
+        self.fisher = fisher
+        Game.__init__(self, clocks, miliseconds)
+    
+    def end_turn(self):
+        """end the turn, fisher style
+
+        this increments the current clock before switching turns as normal
+        """
+        self.cur_clock.time += self.fisher
+        Game.end_turn(self)
+
+class ClassicGame(Game):
+    """a classic board-type game
+
+    every player gets N time to play his turn and his clock is reset at the end of his turn
+    """
+
+    def end_turn(self):
+        """override the end_turn function to reset the timers at the end of turns"""
+        global default_time
+        tmp = self.cur_clock
+        Game.end_turn(self)
+        tmp.time = default_time
+        tmp.update()
+
+class HourglassGame(Game):
+    """hourglass modelisation
+
+    this is like an hourglass on a table that is flipped at the end of each turn
+
+    note that this will probably work very well with > 2 players.
+    """
+
+    def __init__(self, clocks, miliseconds):
+        """this overrides the base constructor to make only one clock have an initial time
+
+        basically, this is to represent that when you start it, the hour glass is empty on one side
+        """
+        Game.__init__(self, clocks, miliseconds)
+        p = self.first_clock.next
+        while p:
+            p.time = 0
+            p = p.next
+
+    def start(self):
+        """reimplement the start() function altogether
+
+        make sure all the clocks are started and they are in the right direction
+        """
+        p = self.first_clock
+        while p:
+            p.start()
+            p.update()
+            # reverse the other clocks
+            if p == self.cur_clock:
+                p.factor = -1
+            else:
+                p.factor = 1
+            p = p.next
+
+    def end_turn(self):
+        """reimplement the end_turn function altogether
+
+        we go to the next clock, stop it, reverse its direction and
+        start it again
+        """
+        self.cur_clock = self.next_clock()
+        p = self.first_clock
+        while p:
+            p.stop()
+            if p == self.cur_clock:
+                p.factor = -1
+            else:
+                p.factor = 1
+            p.start()
+            p = p.next
+        self.turns += 1
 
 class GameclockGtkUI:
     """this class handles most of the UI and turned-based logic
@@ -387,17 +475,21 @@ class GameclockGtkUI:
         label.show()
         type_layout.pack_start(label, True, True, 0)
 
-        button = gtk.RadioButton(None, "Blitz")
-        button.connect("toggled", self.handle_blitz, "Blitz")
-        button.set_active(True)
-        type_layout.pack_start(button, True, True, 0)
-        button.show()
-   	
+        button = None
+        for type in [ "Classic", "Blitz", "Hourglass" ]:
+            button = gtk.RadioButton(button, type)
+            if default_game == "%sGame" % type:
+                button.set_active(True)
+            button.connect("toggled", self.handle_type, type)
+            type_layout.pack_start(button, True, True, 0)
+            button.show()
+
+        # fisher is special: we need to pack more stuff on its line
         fisher_layout = gtk.HBox(False, 10)
         type_layout.pack_start(fisher_layout, True, True, 0)
 
         button = gtk.RadioButton(button, "Fisher")
-        button.connect("toggled", self.handle_fisher, "Fisher")
+        button.connect("toggled", self.handle_type, "Fisher")
         fisher_layout.pack_start(button, True, True, 0)
         button.show()
 
@@ -481,6 +573,15 @@ class GameclockGtkUI:
         
         this handler is ran periodically through a timeout signal to make sure that the current clock is updated
         """
+
+        if isinstance(self.game, HourglassGame):
+            p = self.game.first_clock
+            while p:
+                p.update()
+                p = p.next
+            self.refresh()
+            return True
+
         self.cur_clock.label.set_label(self.game.cur_clock.update())
         active = self.cur_clock.label.get_parent()
         if self.game.cur_clock.dead:
@@ -510,18 +611,39 @@ class GameclockGtkUI:
             p = p.next
             q = q.next
 
-    def setup_game(self, miliseconds = False, fisher = 0):
+    def setup_game(self):
         """initial setup of the clocks, game and labels"""
-        try:
-            players = int(self.players_val.get_value())
-        except:
-            players = default_players
 
-        # the game
-        self.game = Game(players, miliseconds, fisher)
-        self.maybe_print("game setup with %d players" % players)
+        try:
+            # this may be a reset, init from old values
+            miliseconds = self.game.miliseconds
+            name = self.game.__class__.__name__
+        except AttributeError:
+            miliseconds = False
+            # it is here we define the default game
+            name = default_game
+
+        if name == "FisherGame":
+            self.game = FisherGame(self.count_players(), miliseconds, self.get_fisher())
+        else:
+            self.game = eval("%s(%d, %d)" % (name, self.count_players(), miliseconds))
+        self.maybe_print("game setup with %d players" % self.count_players())
         self.refresh()
 
+    def count_players(self):
+        """return the number of players configured"""
+        try:
+            return int(self.players_val.get_value())
+        except:
+            return default_players
+
+    def get_fisher(self):
+        """return the current fisher delay selected in the UI or the default if there is no UI yet"""
+        try:
+            return int(self.fisher_val.get_value() * 1000)
+        except:
+            return default_fisher
+
     def handle_key_press(self, widgets, event):
         keyname = gtk.gdk.keyval_name(event.keyval)
         if keyname == 'Shift_L':
@@ -543,22 +665,18 @@ class GameclockGtkUI:
 
     # XXX: those could be simplified by tagging the radio with the
     # mode, which would be passed along
-    def handle_blitz(self, widget, data=None):
+    def handle_type(self, widget, data=None):
         if widget.get_active():
-            self.game.mode = 'blitz'
-            self.game.fisher = 0
-            self.fisher_val_btn.hide()
-            self.fisher_label.hide()
-            self.maybe_print("using blitz mode")
-        return
-
-    def handle_fisher(self, widget, data=None):
-        if widget.get_active():
-            self.game.mode = 'fisher'
-            self.handle_fisher_change(self.fisher_val)
-            self.fisher_val_btn.show()
-            self.fisher_label.show()
-            self.maybe_print("using fisher mode")
+            if data == "Fisher":
+                # special case: different arguments
+                self.game = FisherGame(self.count_players(), self.game.miliseconds, self.get_fisher())
+                self.fisher_val_btn.show()
+                self.fisher_label.show()
+            else:
+                self.game = eval("%sGame(%d, %d)" % (data, self.count_players(), self.game.miliseconds))
+                self.fisher_val_btn.hide()
+                self.fisher_label.hide()
+            self.maybe_print("using %s mode" % data)
         return
 
     def handle_fisher_change(self, widget, data=None):
@@ -598,8 +716,11 @@ class GameclockGtkUI:
             p = p.next
             q = q.next
             count += 1
+        cols = 2
+        rows = (count - 1)/ cols + 1
+        self.clock_table.resize(rows, cols)
         self.refresh()
-        self.maybe_print("now at %d clocks" % count)
+        self.maybe_print("now at %d clocks, table size is %dX%d" % (count, rows, cols))
 
     def handle_end_turn(self, widget = None, data=None):
         """handle end turn events
@@ -677,10 +798,7 @@ class GameclockGtkUI:
         * show the controls and fix the hilight
         """
         self.maybe_print("reseting game")
-        ms = self.game.miliseconds
-        fisher = self.game.fisher
-        del self.game
-        self.setup_game(ms, fisher)
+        self.setup_game()
         self.cur_clock = self.first_clock
         try:
             gobject.source_remove(self.timeout_source) # stop the refresh
@@ -695,15 +813,16 @@ class GameclockGtkUI:
         self.hilight()
 
     def handle_clock_change(self, widget = None):
+        global default_time
         hours = self.hours_val.get_value()
         minutes = self.minutes_val.get_value()
         seconds = self.seconds_val.get_value()
+        default_time = ( ( hours * 60 * 60 ) + ( minutes * 60 ) + seconds ) * 1000
 
         # XXX: note that we edit all clocks at the same time
         p = self.game.first_clock
         while p:
-            p.time = ( hours * 60 * 60 ) + ( minutes * 60 ) + seconds
-            p.time *= 1000
+            p.time = default_time
             p.update()
             p = p.next
         self.refresh()
@@ -723,7 +842,7 @@ class GameclockGtkUI:
         self.maybe_print("activating miliseconds display")
 
 def usage():
-    print """gameclock v%s  Copyright (C) 2008  Antoine Beaupré
+    """gameclock v%s  Copyright (C) 2008  Antoine Beaupré
 This program comes with ABSOLUTELY NO WARRANTY.
 This is free software, and you are welcome to redistribute it
 under certain conditions.
@@ -738,7 +857,8 @@ Usage:
   -t --time <n>: start time in seconds
   -p --players <n>: start n clocks
 
-See the manpage for more information.""" % (version, sys.argv[0])
+See the manpage for more information."""
+    print usage.__doc__ % (version, sys.argv[0])
 
 if __name__ == "__main__":
     try:
